
if __package__=="eznet_keras.models":
    from ..utils import *
else:
    import os, sys
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    sys.path.append(parent_dir)
    from utils import *

class KerasSmartModel(tf.keras.models.Model):
    
    sample_hparams = {
        'model_name': 'KerasSmartModel',
        'l2_reg': 0.0001,
        'batch_size': 16,
        'epochs': 2,
        'validation_data': 0.1,
        'validation_tolerance_epochs': 10,
        'learning_rate': 0.0001,
        'learning_rate_decay_gamma': 0.99,
        'loss_function': 'categorical_crossentropy',
        'loss_function_params': None,
        'metrics':['accuracy'],
        'optimizer': 'Adam',
        'optimizer_params': None,
        'checkpoint_path':None,
        'early_stopping_monitor':'loss',
        'early_stopping_mode':'min',
        'early_stopping_value':1.0e-6
    }
    
    def __init__(self, hparams:dict=None):
        """
        Base class for smart, trainable pytorch modules. All hyperparameters are contained within the `hparams`
        dictionary. Some training-related hyperparameters are common across almost all kinds of PyTorch modules,
        which can be overloaded by the child class. The module includes functions for training, evaluation, and
        prediction. These functions cane be modified or overloaded by any child subclass.

        ### Usage

        `net = KerasSmartModel(hparams)` where `hparams` is dictionary of hyperparameters containing the following:
            - `model_name` (str): Name of the model.
            - `batch_size` (int): Minibatch size, the expected input size of the network.
            - `learning_rate` (float): Initial learning rate of training.
            - `learning_rate_decay_gamma` (float): Exponential decay rate gamma for learning rate, if any.
            - `optimizer` (str): Optimizer. Examples are "Adam", "SGD", "RMSprop", "Adagrad", etc.
            - `optimizer_params` (dict): Additional parameters of the optimizer, if any.
            - `epochs` (int): Maximum number of epochs for training.
            - `validation_tolerance_epochs` (int): Epochs to tolerate unimproved val loss, before early stopping (patience).
            - `l2_reg` (float): L2 regularization parameter.
            - `loss_function` (str): Loss function. Examples: "MSELoss", "CrossEntropyLoss", "NLLLoss", etc.
            - `loss_function_params` (dict): Additional parameters for the loss function, if any.
            - `metrics` (list): list of metrics for Keras compilation.
            - `checkpoint_path` (str): Path to the directory where checkpoints will be saved at every epoch.
            - `early_stopping_monitor` (str): Monitor whose critical value will cause early stopping. Default is 'loss', but 'val_loss' is typically used.
            - `early_stopping_mode` (str): Mode of the parameter whose critical value will be used for early stopping. Deafults to 'min' for any error. 'max' is for accuracy, etc.
            - `early_stopping_value` (float): Value of the monitor at which point training will stop becasue the critical value has been reached.

        ### Returns

        Returns a `tf.keras.models.Model` object that can be trained and used accordingly.
        Run `net.summary()` afterwards to see what you have inside the network.
        
        ### Notes:
        - `self.batch_input_shape` attribute must be set in the `__init__` method.
        - `self.batch_output_shape` attribute must be set in the `__init__` method.
        - `self._callbacks` is a list of callbacks sent to the training method of Keras. It already includes garbage collection.
        - `self._es` is an EarlyStopping instance of Keras, if specified, otherwise None.
        - `self._chkpt` is the hyperparameter `checkpoint_path` from the input dictionary, if it exists, otherwise None.
        - `self._chk` is a ModelCheckpoint instance sent to the training function of Keras, if specified, otherwise None.
        - `self._es_crit` is an EarlyStopAtCriteria instance that stops the training if 'val_loss' for instance, reaches a certain value.
        - `self.net` **MUST** always exist so that Keras2Cpp can serialize the layers that it understands.
          `self.net` is always used as the model within this module, that contains the network itself. It is typically a Sequential or Functional model built in the `__init__` 
          method.
        """
        super(KerasSmartModel, self).__init__()
        if not hparams: hparams = self.sample_hparams
        self.hparams = hparams
        self._batch_size = int(hparams["batch_size"])
        self._loss_function = hparams.get("loss_function")
        self._loss_function_params = hparams.get("loss_function_params")
        self._metrics_list = hparams.get("metrics")
        self._optimizer = hparams.get("optimizer")
        self._optimizer_params = hparams.get("optimizer_params")
        self._validation_tolerance_epochs = hparams.get("validation_tolerance_epochs")
        self._learning_rate = hparams.get("learning_rate")
        self._learning_rate_decay_gamma = hparams.get("learning_rate_decay_gamma")
        self._validation_data = hparams.get("validation_data")
        self._epochs = hparams.get("epochs")
        self._l2_reg = hparams.get("l2_reg") if hparams.get("l2_reg") else 0.0
        self.history = None
        self.batch_input_shape = (self._batch_size, 1)
        self.batch_output_shape = (self._batch_size, 1)
        self._callbacks = [GarbageCollectionCallback()]
        self._es = tf.keras.callbacks.EarlyStopping(monitor='val_loss', mode="min", patience=self._validation_tolerance_epochs) if self._validation_tolerance_epochs else None
        if self._es:
            self._callbacks.append(self._es)
        self._chkpt = hparams.get("checkpoint_path")
        if self._chkpt:
            self._chk = tf.keras.callbacks.ModelCheckpoint(self._chkpt, monitor='val_loss', verbose=0, save_best_only=True, mode='min')
        else:
            self._chk = None
        if self._chk:
            self._callbacks.append(self._chk)
        self.early_stopping_monitor = hparams.get("early_stopping_monitor")
        self.early_stopping_mode = hparams.get("early_stopping_mode")
        self.early_stopping_value = hparams.get("early_stopping_value")
        if self.early_stopping_monitor:
            self._es_crit = EarlyStopAtCriteria(monitor=self.early_stopping_monitor, mode=self.early_stopping_mode, value=self.early_stopping_value)
            self._callbacks.append(self._es_crit)
        else:
            self._es_crit = None
        self.net = tf.keras.models.Sequential()
    
    
    def call(self, x, *args, **kwargs):
        return self.net(x, *args, **kwargs)
    
    def build(self):
        super().build(input_shape=self.batch_input_shape) 
        
    def summary(self):
        return self.net.summary()
        
    def get_config(self):
        config = super(KerasSmartModel, self).get_config()
        config['hparams'] = self.hparams
        return config
    
    @classmethod
    def from_config(cls, config):
        return cls(config['hparams'])
    
    def compile(self, num_samples=None):
        compile_keras_model(self.net, self._batch_size, self._learning_rate, self._optimizer, self._loss_function, 
                              self._metrics_list, self._optimizer_params, self._learning_rate_decay_gamma, num_samples)
            
    def fit(self, x_train, y_train, x_val, y_val, verbose:bool=True, **kwargs):
        self.history = fit_keras_model(self.net, x_train, y_train, x_val, y_val, 
            self._batch_size, self._epochs, self._callbacks, verbose, **kwargs)
        return self.history

    def __str__(self):
        s = "KerasSmartModel with the following attributes:\n" + str(self.hparams)
        return s
    
    def train(self, x_train, x_val, y_train, y_val, verbose:bool=True, saveto:str=None, export:str=None, **kwargs):
        """Train the model according to its hyperparameters.

        ### Args:
            - `x_train` (numpy array): Training inputs
            - `x_val` (numpy array): Validation inputs
            - `y_train` (numpy array): Training target outputs
            - `y_val` (numpy array): Validation target outputs
            - `verbose` (bool, optional): Verbosity of training. Defaults to True.
            - `saveto` (str, optional): Save Keras model in path. Defaults to None.
            - `export` (str, optional): Save Keras model in .model file using keras2cpp for later use in C++. Defaults to None.

        ### Returns:
            Nothing. It modifies the "net" attribute of the model, and the history of the training in self.history.
        
        """
        N = x_train.shape[0]
        self.compile(num_samples=N)
        _ = self.fit(x_train, y_train, x_val, y_val, verbose=verbose, **kwargs)
        if saveto:
            save_keras_model(self.net, self.history.history, saveto, self.hparams)
        if export:
            export_keras_model(self.net, export)

    def plot_history(self, metrics=['loss','val_loss'], fig_title='model loss', saveto:str=None):
        plot_keras_model_history(self.history.history, metrics, fig_title, saveto)
        
    def evaluate(self, *args, **kwargs):
        return self.net.evaluate(*args, **kwargs)
    
    def predict(self, *args, **kwargs):
        return self.net.predict(*args, **kwargs)




if __package__=="eznet_keras.models":
    from ..utils import *
    from .keras_smart_module import *
    from .dense_block import *
else:
    import os, sys
    current_dir = os.path.dirname(os.path.abspath(__file__))
    parent_dir = os.path.dirname(current_dir)
    sys.path.append(parent_dir)
    sys.path.append(current_dir)
    from utils import *
    from keras_smart_module import *
    from dense_block import *


class Recurrent_Network(KerasSmartModel):
    
    sample_hparams = {
        'model_name': 'Recurrent_Network',
        'in_features': 10,
        'out_features': 3,
        'in_seq_len': 13,
        'out_seq_len': 1,
        'rnn_type': 'LSTM',
        'rnn_hidden_sizes': 8,
        'rnn_bidirectional': False,
        'rnn_depth': 2,
        'rnn_dropout': 0.1,
        'rnn_params': None,
        'final_rnn_return_sequences': False,
        'apply_dense_for_each_time_step': True,
        'permute_output': False,
        'dense_width': 16,
        'dense_depth': 2,
        'dense_dropout': 0.2,
        'dense_activation': 'relu',
        'dense_activation_params': None,
        'output_activation': None,
        'output_activation_params': None,
        'norm_layer_type': 'BatchNormalization',
        'norm_layer_params': None,
        'norm_layer_position': 'before',
        'l2_reg': 0.0001,
        'batch_size': 16,
        'epochs': 2,
        'validation_data': [0.05,'testset'],
        'validation_tolerance_epochs': 10,
        'learning_rate': 0.0001,
        'learning_rate_decay_gamma': 0.99,
        'loss_function': 'categorical_crossentropy',
        'loss_function_params': None,
        'metrics':['accuracy'],
        'optimizer': 'Adam',
        'optimizer_params': None,
        'checkpoint_path':None,
        'early_stopping_monitor':'loss',
        'early_stopping_mode':'min',
        'early_stopping_value':1.0e-6
    }
    
    def __init__(self, hparams:dict=None):
        """Sequence to Dense network with RNN for time-series classification, regression, and forecasting, as well as NLP applications.
        This network uses any RNN layers as encoders to extract information from input sequences, and fully-connected 
        multilayer perceptrons (Dense) to decode the sequence into an output, which can be class probabilitites 
        (timeseries classification), a continuous number (regression), or an unfolded sequence (forecasting) of a 
        target timeseries.

        ### Usage

        `net = Recurrent_Network(hparams)` where `hparams` is dictionary of hyperparameters containing the following:
        
            - `model_name` (str): Name of the model, can be used later for saving, etc.
            - `rnn_type` (str): RNN type, options are "LSTM", "GRU", "SimpleRNN", etc.
            - `in_seq_len` (int): Input sequence length, in number of timesteps
            - `out_seq_len` (int): Output sequence length, in number of timesteps, assuming output is also a sequence. This will affect the output layer in the dense section.
                Use 1 for when the output is not a sequence, or do not supply this key.
            - `in_features` (int): Number of features of the input.
            - `out_features` (int): Number of features of the output.
            - `rnn_hidden_sizes` (int): RNN layer hidden size. A number sets them all the same. Default is 16.
            - `rnn_bidirectional` (bool): Whether the RNN layers are bidirectional or not. Default is False.
            - `rnn_depth` (int): Number of stacked RNN layers. Default is 1.
            - `rnn_dropout` (float): Dropout rates, if any, of the RNN layers. 
                Please note that using dropout in RNN layers is generally discouraged, for it decreases determinism during inference.
            - `rnn_params` (dict): Dictionary of kwargs for the RNN layer constructor. Default is None.
              If specified, the keys in this dictionary will not only add to, but also overwrite any existing arguments this class passes to the RNN layer constructor.
            - `final_rnn_return_sequences` (bool): Whether the final RNN layer returns sequences of hidden state. 
                **NOTE** Setting this to True will make the model much, much larger.
            - `apply_dense_for_each_time_step` (bool): Whether to apply the Dense network to each time step of the 
                RNN output. If False, the Dense network is applied to the last time step only if 
                `final_rnn_retrurn_sequences` is False, or applied to a flattened version of the output sequence
                otherwise (the dimensionality of the input feature space to the dense network will be multiplied
                by the sequence length. PLEASE NOTE that this only works if the entered sequence is exactly as long
                as the priorly defined sequence length according to the hyperparameters).
            - `permute_output` (bool): Whether to permute the output sequence to be (N, D*H_out, L_out)
            - `dense_width` (int|list): (list of) Widths of the Dense network. It can be a number (for all) or a list holding width of each hidden layer.
            - `dense_depth` (int): Depth (number of hidden layers) of the Dense network.
            - `dense_activation` (str|list): (list of) Activation functions for hidden layers of the Dense network. Examples: "relu", "leaklyrelu", "sigmoid", "tanh", etc.
            - `dense_activation_params` (dict|list): (list of) Dictionaries of parameters for the activation func constructors of the Dense network.
            - `output_activation` (str): Activation function for the output layer of the Dense network, if any. Examples: "softmax", "sigmoid", etc.
                **NOTE** If the loss function is `sparse_categorical_crossentropy`, then no output activation is erquired.
                However, if the loss function is `categorical_crossentropy` (a.k.a. negative loglikelihood), then you must specify an output activation as in "softmax".
            - `output_activation_params` (dict): Dictionary of parameters for the activation func constructor of the output layer.
            - `norm_layer_type` (str|list): (list of) Types of normalization layers to use in the dense section, if any. Options are "BatchNormalization", "LayerNormalization",etc.
            - `norm_layer_params` (dict|list): (list of) Dictionaries of parameters for the normalization layer constructors.
            - `norm_layer_position` (str|list): (list of) Whether the normalization layer should come 'before' or 'after' the activation of each hidden layer in the dense network.
            - `dense_dropout` (float|list): (list of) Dropout rates (if any) for the hidden layers of the Dense network.
            - `batch_size` (int): Minibatch size, the expected input size of the network.
            - `learning_rate` (float): Initial learning rate of training.
            - `learning_rate_decay_gamma` (float): Exponential decay rate gamma for learning rate, if any.
            - `optimizer` (str): Optimizer. Examples: 'Adam', 'SGD', 'RMSProp', etc.
            - `optimizer_params` (dict): Additional parameters of the optimizer, if any.
            - `epochs` (int): Maximum number of epochs for training.
            - `validation_tolerance_epochs` (int): Epochs to tolerate unimproved val loss, before early stopping.
            - `validation_data` (list): Portion of validation data. Should be a tuple like [validation split, dataset as in 'trainset' or 'testset']
            - `l2_reg` (float): L2 regularization parameter.
            - `loss_function` (str): Loss function. Examples: 'binary_crossentropy', 'categorical_crossentropy', 'mse', etc.
            - `loss_function_params` (dict): Additional parameters for the loss function, if any.
            - `metrics` (list): list of metrics for Keras compilation.
            - `checkpoint_path` (str): Path to the directory where checkpoints will be saved at every epoch.
            - `early_stopping_monitor` (str): Monitor whose critical value will cause early stopping. Default is 'loss', but 'val_loss' is typically used.
            - `early_stopping_mode` (str): Mode of the parameter whose critical value will be used for early stopping. Deafults to 'min' for any error. 'max' is for accuracy, etc.
            - `early_stopping_value` (float): Value of the monitor at which point training will stop becasue the critical value has been reached.

        ### Returns
        
        Returns a `tf.keras.models.Model` object that can be trained and used accordingly.
        Run `net.summary()` afterwards to see what you have inside the network.
        The returned model is an instance of `KerasSmartModel`, which has built-in functions for training, evaluation, etc.
        """
        super(Recurrent_Network, self).__init__(hparams)
        hparams = hparams if hparams is not None else self.sample_hparams
        self._rnn_type = hparams["rnn_type"]
        self._rnn = getattr(tf.keras.layers, self._rnn_type)
        self._denseactivation = hparams["dense_activation"] if hparams.get("dense_activation") else "relu"
        self._denseactivation_params = hparams.get("dense_activation_params")
        self._outactivation = hparams.get("output_activation")
        self._outactivation_params = hparams.get("output_activation_params")
        self._normlayer_type = hparams.get("norm_layer_type")
        self._normlayer_params = hparams.get("norm_layer_params")
        self._normlayer_position = hparams.get("norm_layer_position")
        self._infeatures = hparams["in_features"]
        self._outfeatures = hparams["out_features"]
        self._rnnhidsizes = hparams["rnn_hidden_sizes"] if hparams.get("rnn_hidden_sizes") else 16
        self._densehidsizes = hparams["dense_width"] if hparams.get("dense_width") else 16
        self._densedepth = hparams["dense_depth"] if hparams.get("dense_depth") else 0
        self._rnndepth = hparams["rnn_depth"] if hparams.get("rnn_depth") else 1
        self._bidirectional = True if hparams.get("rnn_bidirectional") else False
        self._rnndropout = hparams["rnn_dropout"] if hparams.get("rnn_dropout") else 0
        self._densedropout = hparams["dense_dropout"] if hparams.get("dense_dropout") else None
        self._final_rnn_return_sequences = True if hparams.get("final_rnn_return_sequences") else False
        self._apply_dense_for_each_timestep = True if hparams.get("apply_dense_for_each_timestep") else False
        self._permute_output = True if hparams.get("permute_output") else False
        self._N = int(hparams["batch_size"])
        self._L_in = int(hparams["in_seq_len"])
        self._L_out = int(hparams["out_seq_len"]) if hparams.get("out_seq_len") else 1
        self._D = int(2 if self._bidirectional else 1)
        self._rnn_params = hparams.get("rnn_params")
        self._H_in = int(self._infeatures)
        if self._rnnhidsizes == "auto": self._H_cell = int(2**(np.round(math.log2(self._H_in*self._L_in))))
        else: self._H_cell = int(self._rnnhidsizes)
        self._H_out = int(self._H_cell)
        self.batch_input_shape = (self._N, self._L_in, self._H_in)
        if self._final_rnn_return_sequences and self._apply_dense_for_each_timestep:
            if self._permute_output: self.batch_output_shape = (self._N, self._outfeatures, self._L_out)
            else: self.batch_output_shape = (self._N, self._L_out, self._outfeatures)
        else: self.batch_output_shape = (self._N, self._L_out * self._outfeatures)
        self._out_activation_module = actdict_keras[self._outactivation] if self._outactivation else None
        
        # Initializing sequential network
        self.net = tf.keras.models.Sequential()
        
        # Constructing RNN layers
        _kwargs = {'units':self._H_cell, 'dropout':self._rnndropout, 
            'return_sequences':(True if self._rnndepth > 1 else self._final_rnn_return_sequences), 
            'kernel_regularizer':(tf.keras.regularizers.L2(self._l2_reg) if self._l2_reg else None)}
        if self._rnn_params: 
            _kwargs.update(self._rnn_params)
        self.net.add(tf.keras.Input((self._L_in,self._H_in)))
        for i in range(self._rnndepth):
            if i != self._rnndepth-1:
                _kwargs.update({'return_sequences':True})
            else:
                _kwargs.update({'return_sequences':self._final_rnn_return_sequences})
            self.net.add(tf.keras.layers.Bidirectional(self._rnn(**_kwargs)) if self._bidirectional else self._rnn(**_kwargs))
        
        # If the final RNN layer returns sequences, and we are NOT applying the dense on each time step of it, then the returned sequence must be flattened before being decocded.
        if self._final_rnn_return_sequences and not self._apply_dense_for_each_timestep:
            self.net.add(tf.keras.layers.Flatten())
        
        # Calculating Dense layers widths
        cf = self._L_in if (self._final_rnn_return_sequences and not self._apply_dense_for_each_timestep) else 1 
        self._dense_input_size = self._H_out * self._D * cf
        if self._final_rnn_return_sequences and not self._apply_dense_for_each_timestep:
            self._dense_output_size = int(self._L_out*self._outfeatures)
        else:
            self._dense_output_size = int(self._outfeatures)
            
        # Generate arrays containing parameters of each Dense Block (Every block contains a linear, normalization, activation, and dropout layer).
        self._dense_width_vec = self._gen_hparam_vec_for_dense(self._densehidsizes, 'dense_width')
        self._dense_activation_vec = self._gen_hparam_vec_for_dense(self._denseactivation, 'dense_activation')
        self._dense_activation_params_vec = self._gen_hparam_vec_for_dense(self._denseactivation_params, 'dense_activation_params')
        self._dense_norm_layer_type_vec = self._gen_hparam_vec_for_dense(self._normlayer_type, 'norm_layer_type')
        self._dense_norm_layer_params_vec = self._gen_hparam_vec_for_dense(self._normlayer_params, 'norm_layer_params')
        self._dense_norm_layer_position_vec = self._gen_hparam_vec_for_dense(self._normlayer_position, 'norm_layer_position')
        self._dense_dropout_vec = self._gen_hparam_vec_for_dense(self._densedropout, 'dense_dropout')
        
        # Construct the dense layers
        in_size = self._dense_input_size
        for i in range(self._densedepth):
            out_size = self._dense_width_vec[i]
            temp_dropout_rate = self._dense_dropout_vec[i] if (i != self._densedepth-1) else None # The hidden layer just before the output layer rarely has Dropout.
            add_dense_block(self.net, output_size=out_size, input_shape=None, activation=self._dense_activation_vec[i], activation_params=self._dense_activation_params_vec[i], 
                    norm_layer_type=self._dense_norm_layer_type_vec[i], norm_layer_position=self._dense_norm_layer_position_vec[i], 
                    norm_layer_params=self._dense_norm_layer_params_vec[i], dropout=temp_dropout_rate, 
                    kernel_regularizer=(tf.keras.regularizers.L2(self._l2_reg) if self._l2_reg else None))
            in_size = out_size
        
        # Output layer
        self.net.add(tf.keras.layers.Dense(self._dense_output_size))
        if self._outactivation:
            if self._outactivation_params: self.net.add(tf.keras.layers.Activation(self._out_activation_module(**self._outactivation_params)))
            else: self.net.add(tf.keras.layers.Activation(self._out_activation_module))
            
        # Permute if necessary
        if self._final_rnn_return_sequences and self._apply_dense_for_each_timestep and self._permute_output:
            self.net.add(tf.keras.layers.Permute((2,1)))
        
    def _gen_hparam_vec_for_dense(self, hparam, hparam_name, **kwargs):
        return generate_array_for_hparam(hparam, self._densedepth, hparam_name=hparam_name, count_if_not_list_name='dense_depth', **kwargs)
    
    def call(self, x, *args, **kwargs):
        return self.net(x, *args, **kwargs)



if __name__ == '__main__':
    hparams = Recurrent_Network.sample_hparams
    hparams['rnn_bidirectional'] = True
    model = Recurrent_Network(hparams)
    model.summary()



# Global Constants
SEED = 42

# General-Purpose Libraries
if __name__ == "eznet_keras.utils":
    from .keras2cpp import export_model
else:
    from keras2cpp import export_model
from pathlib import Path
import math
import json
import random
random.seed(SEED)
import numpy as np
np.random.seed(SEED)
from timeit import default_timer as timer
from datetime import datetime
from matplotlib import pyplot as plt

# Tensorflow Libraries
import tensorflow as tf
tf.random.set_seed(SEED)
import gc

# Reset seeds, just in case they were modified during imports
random.seed(SEED)
np.random.seed(SEED)
tf.random.set_seed(SEED)
GLOROTUNIFORM = tf.keras.initializers.GlorotUniform(seed=SEED)
ORTHOGONAL = tf.keras.initializers.Orthogonal(seed=SEED)


########################################################################################################################
# Global variables, functions, and classes
########################################################################################################################

optdict_keras = {'adam':tf.keras.optimizers.Adam, 'sgd':tf.keras.optimizers.SGD, 'rmsprop':tf.keras.optimizers.RMSprop, 'adagrad':tf.keras.optimizers.Adagrad}
actdict_keras = {
    'relu':tf.keras.activations.relu, 'leakyrelu':tf.keras.layers.LeakyReLU, 
    'sigmoid':tf.keras.activations.sigmoid, 'tanh':tf.keras.activations.tanh, 'softmax':tf.keras.activations.softmax,
    'softplus':tf.keras.activations.softplus, 'softsign':tf.keras.activations.softsign,
    'elu':tf.keras.activations.elu, 'selu':tf.keras.activations.selu}
rnndict_keras = {'LSTM':tf.keras.layers.LSTM, 'GRU':tf.keras.layers.GRU, 'SimpleRNN':tf.keras.layers.SimpleRNN}
convdict_keras = {"conv1d":tf.keras.layers.Conv1D, "conv2d":tf.keras.layers.Conv2D, "conv3d":tf.keras.layers.Conv3D}


def make_path(path:str):
    Path.mkdir(Path(path).parent, parents=True, exist_ok=True)
    return path
    

def plot_keras_model_history(history:dict, metrics:list=['loss','val_loss'], 
                             fig_title:str='model loss', saveto:str=None):
    plt.figure(figsize=(15, 7))
    plt.grid(True)
    plt.plot(history[metrics[0]])
    plt.plot(history[metrics[1]])
    plt.title(fig_title)
    plt.ylabel(metrics[0])
    plt.xlabel('epoch')
    plt.legend(['train', 'val'], loc='upper left')
    plt.show()
    if saveto:
        plt.savefig(make_path(saveto), dpi=600)
        

def compile_keras_model(model, _batchsize:int, _learnrate:float, _optimizer:str, _loss:str, _metrics:list, 
                          _optimizerparams:dict=None, _learnrate_decay_gamma:float=None, num_samples:int=None):
    if _learnrate_decay_gamma:
        itersPerEpoch = (num_samples//_batchsize) if num_samples else 1
        sch = tf.keras.optimizers.schedules.ExponentialDecay(initial_learning_rate=_learnrate, 
        decay_steps=itersPerEpoch, decay_rate=_learnrate_decay_gamma)
        lr = sch
    else:
        lr = _learnrate
    if _optimizerparams:
        optparam = _optimizerparams
        # opt = optdict_keras[_optimizer](learning_rate=lr, **optparam)
        opt = getattr(tf.keras.optimizers, _optimizer)(learning_rate=lr, **optparam)
    else:
        # opt = optdict_keras[_optimizer](learning_rate=lr)
        opt = getattr(tf.keras.optimizers, _optimizer)(learning_rate=lr)
    model.compile(optimizer=opt, loss=_loss, metrics=_metrics)
    

def fit_keras_model(model, x_train, y_train, x_val, y_val, 
    _batchsize:int, _epochs:int, _callbacks:list, verbose:bool=True, **kwargs):
    while True:
        try:
            history = model.fit(x_train, y_train, batch_size=_batchsize, epochs=_epochs, 
                validation_data=(x_val, y_val), verbose=verbose, 
                callbacks=_callbacks, **kwargs)
            break
        except Exception as e:
            print(e)
            print(("\nTraining failed with batchsize={}. "+\
                "Trying again with a lower batchsize...").format(_batchsize))
            _batchsize = _batchsize // 2
            if _batchsize < 2:
                raise ValueError("Batchsize too small. Training failed.")
    return history


def save_keras_model(model, history:dict, path:str, hparams:dict):
    try:
        model.save(make_path(path))
        for key in history:
            hparams[key] = history[key]
        jstr = json.dumps(hparams, indent=4)
        with open(path+"/hparams.json", "w") as f:
            f.write(jstr)
    except Exception as e:
        print(e)
        print("Cannot serialize Keras model.")
        
        
def export_keras_model(model, path:str):
    try:
        export_model(model, make_path(path))
        print("Model exported successfully.")
    except Exception as e1:
        print(e1)
        print("Cannot export Keras model using keras2cpp on the fly. Will try sequentializing the model layers...")
        try:
            net = tf.keras.models.Sequential(model.layers)
            export_model(net, make_path(path))
            print("Model exported successfully.")
        except Exception as e2:
            print(e2)
            print("Cannot export model using Keras2Cpp.")
    

def test_keras_model_class(model_class):
    print("Constructing model...\n")
    model = model_class()
    print("Summary of model:")
    print(model.summary())
    print("\nGenerating random dataset...\n")
    (x_train, y_train) = generate_sample_batch(model)
    (x_val, y_val) = generate_sample_batch(model)
    print("\nTraining model...\n")
    model.train(x_train, x_val, y_train, y_val, 
                verbose=True, saveto="test_"+model.hparams["model_name"], 
                export="test_"+model.hparams["model_name"]+".model")
    print("\nEvaluating model...\n")
    model.evaluate(x_val, y_val, verbose=True)
    print("Done.")
    


# Perform garbage collection
class GarbageCollectionCallback(tf.keras.callbacks.Callback):
    def on_epoch_end(self, epoch, logs=None):
        gc.collect()
        
        
# Deploy early stopping when performance reaches good values
class EarlyStopAtCriteria(tf.keras.callbacks.Callback):
    def __init__(self, monitor='val_loss', mode='min', value=0.001):
        super(EarlyStopAtCriteria, self).__init__()
        self.monitor = monitor
        self.value = value
        self.mode = mode
    def on_epoch_end(self, epoch, logs=None):
        if self.mode == 'min':
            if logs.get(self.monitor) <= self.value:
                print("Early stopping performance criteria has been reached. Stopping training.")
                self.model.stop_training = True
        else:
            if logs.get(self.monitor) >= self.value:
                print("Early stopping performance criteria has been reached. Stopping training.")
                self.model.stop_training = True
        

# Sampling layer used for variational autoencoders
class Sampling(tf.keras.layers.Layer):
    """Uses (z_mean, z_log_var) to sample z, the vector encoding a digit."""

    def call(self, inputs):
        z_mean, z_log_var = inputs
        batch = tf.shape(z_mean)[0]
        dim = tf.shape(z_mean)[1]
        epsilon = tf.keras.backend.random_normal(shape=(batch, dim))
        return z_mean + tf.exp(0.5 * z_log_var) * epsilon


def autoname(name):
    """
    Genereate a unique name for a file, based on the current time and the given name.
    Gets the `name` as a string and adds the time stamp to the end of it before returning it.
    """
    return name + "_" + datetime.now().strftime("%Y_%m_%d_%H_%M_%S")



def calc_image_size(size_in:int, kernel_size:int, padding:int, stride:int, dilation:int):
    """Calculate image size after convolution or pooling.
    
    **NOTE** grouping is not supported yet.

    ### Args:
        - `size_in` (int|list): (list of) image input dimension(s).
        - `kernel_size` (int|list): (list of) kernel (or pool) size
        - `padding` (int|list): (list of) padding sizes, or string such as 'valid' and 'same'.
        - `stride` (int|list): (list of) strides.
        - `dilation` (int|list): (list of) dilation rates.

    ### Returns:
        int or list: Output image dimensions
    """
    if padding == 'same':
        return size_in
    else:
        if padding == 'valid':
            padding = 0
        if isinstance(size_in, (list, tuple)):
            if isinstance(padding, int): padding = [padding]*len(size_in)
            if isinstance(kernel_size, int): kernel_size = [kernel_size]*len(size_in)
            if isinstance(stride, int): stride = [stride]*len(size_in)
            if isinstance(dilation, int): dilation = [dilation]*len(size_in)
            return [math.floor((size_in[i] + 2*padding[i] - dilation[i]*(kernel_size[i]-1) - 1)/stride[i] + 1) for i in range(len(size_in))]
        else:
            assert isinstance(size_in, int), "size_in must be an integer or a list/tuple of integers."
            assert isinstance(padding, int), "padding must be an integer or a list/tuple of integers."
            assert isinstance(kernel_size, int), "kernel_size must be an integer or a list/tuple of integers."
            assert isinstance(stride, int), "stride must be an integer or a list/tuple of integers."
            assert isinstance(dilation, int), "dilation must be an integer or a list/tuple of integers."
            return math.floor((size_in + 2*padding - dilation*(kernel_size-1) - 1)/stride + 1)


def generate_geometric_array(init, count, direction, powers_of_two=True):
    """Generate array filled with incrementally doubling/halving values, optionally with powers of two.

    ### Args:
        `init` (int): The first value to begin.
        `count` (int): Number of elements to generate.
        `direction` (str): Direction of the array. Can be either 'up' or 'down', i.e. increasing or decreasing.
        `powers_of_two` (bool, optional): Generate numbers that are powers of two. Defaults to True.

    ### Returns:
        list: List containing elements
    """
    lst = []
    old = int(2**math.ceil(math.log2(init))) if powers_of_two else init
    new = old
    for _ in range(count):
        lst.append(new)
        old = new
        new = (old * 2) if direction == 'up' else (old // 2)
    return lst


def generate_array_for_hparam(
    hparam, count_if_not_list:int, 
    hparam_name:str='parameter', count_if_not_list_name:str='its count',
    check_auto:bool=False, init_for_auto:int=2, powers_of_two_if_auto:bool=True,
    direction_if_auto:str=None):
    """Generate array for a hyperparameter, regardless of if it is a list or not. This function is for use in APIs
    that generate models with hyperparameters as inputs, which can be lists, a single item, or "auto".
    Examples include width of a neural network's hidden layers, channels of conv layers, etc.
    For these hyperparameters, the user is typically free to specify an array-like, a single item to be repeated,
    or "auto" for automatic calculation of the parameter.
    This function is meant to be used in the body of the code of class constructors and other functions in the API.

    ### Args:
        `hparam` (var): A specific hyperparameter, e.g., input by user to your network's constructor.
        `count_if_not_list` (int): Number of elements to generate if `hparam` is not an array-like.
        `hparam_name` (str, optional): Name of the hyperparameter. Defaults to 'parameter'.
        `count_if_not_list_name` (str, optional): Name of the "count" that must be provided. Defaults to 'its count'.
        `check_auto` (bool, optional): Check for the "auto" case. Defaults to False.
        `init_for_auto` (int, optional): Initial value in case of "auto". Defaults to 2.
        `powers_of_two_if_auto` (bool, optional): Generate powers of two in case of "auto". Defaults to True.
        `direction_if_auto` (str, optional): Direction of geometric increment in case of "auto". Defaults to None.
        This can be "up" or "down". If check_for_auto is True, then this argument must be specified.

    ### Returns:
        list: List containing elements
    """
    assert count_if_not_list is not None, \
        "Since %s may not be a list/tuple, %s must always be specified."%(hparam_name, count_if_not_list_name)
    if isinstance(hparam, (list,tuple)) and len(hparam) == count_if_not_list:
        lst = hparam
    elif hparam == "auto" and check_auto:
        assert init_for_auto is not None, \
            "If %s is 'auto', then %s must be specified."%(hparam_name, "init_for_auto")
        assert direction_if_auto in ['up','down'], \
            "If %s is 'auto', then %s must be specified as 'up' or 'down'."%(hparam_name, "direction_if_auto")
        lst = generate_geometric_array(init_for_auto, count_if_not_list, direction_if_auto, powers_of_two_if_auto)
    else:
        lst = [hparam]*count_if_not_list
    return lst




def generate_sample_batch(model):
    x = np.random.rand(*model.batch_input_shape).astype(np.float32)
    y = np.random.rand(*model.batch_output_shape).astype(np.float32)
    if model.hparams['loss_function'] in ['sparse_categorical_crossentropy']:
        y = np.argmax(y, axis=1)
    return (x,y)
    


if __package__=="eznet_keras.models":
    from ..utils import *
else:
    import os, sys
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    sys.path.append(parent_dir)
    from utils import *
import warnings

class Conv_Block(tf.keras.layers.Layer):
    def __init__(self, input_shape:list, out_channels:int=None, conv_dim:int=1, input_image:list=[30], conv_kernel_size=3, conv_padding='valid', conv_stride=1, conv_dilation=1, 
                 conv_params:dict=None, conv_activation:str='relu', conv_activation_params:dict=None, norm_layer_position:str=None, norm_layer_type:str=None, 
                 norm_layer_params:dict=None, pool_type:str=None, pool_kernel_size=2, pool_padding:str='valid', pool_stride=1, pool_params:dict=None, 
                 dropout:float=None, min_image_dim:int=1, kernel_regularizer:tf.keras.regularizers.Regularizer=None):
        """Convolutional block, containing one convolution layer, followed optionally by a normalization layer, an activation layer, a pooling layer and a dropout layer. The
        convolution layer is mandatory, but the other ones are optional. The convolution layer can be 1D, 2D or 3D. The normalization layer can be any such layer defined
        in Keras. The activation layer can also be anything with lower-case strings, and the pooling layer can be any of the pooling layers defined for Keras. 
        The dropout layer is a spatial dropout layer. The dimension of any dropout layer will match the dimension of the convolution layer. For `Conv1D` for instance, 
        `SpatialDropout1D` will be used, if desired.
        
        This module is meant to be used as a building block for larger modules.

        ### Args:
        
        - `input_shape` (list|tuple): Input shape.
        - `out_channels` (int, optional): Number of convolution filters. Defaults to the input channels size.
        - `conv_dim` (int, optional): Dimension of the convolution. Defaults to 1. 1 means Conv1d, 2 means Conv2d etc.
        - `input_image` (list, optional): Size of the input image. Defaults to [30]. This must be a list/tuple of integers, with legth equal to `conv_dim`.
        - `conv_kernel_size` (int, optional): Convolution kernel size. Defaults to 3. It is strongly recommended to provide a list of integers, with length equal to `conv_dim`.
        - `conv_padding` (str, optional): Convolution padding. Defaults to 'same'. Arrays are recommended over integers.
        - `conv_stride` (int, optional): Convolution stride. Defaults to 1.
        - `conv_dilation` (int, optional): Convolution dilation. Defaults to 1.
        - `conv_params` (dict, optional): Additional dictionary of kwargs for Conv?d module, to add or overwrite other parameters. Defaults to None.
          No matter what parameters are set by this class by default for Con?D layer, providing this dictionary will overwrite all of them.
        - `conv_activation` (str, optional): String representing activation function. Defaults to 'relu'. Examples: 'leakyrelu', 'sigmoid', 'tanh' etc.
        - `conv_activation_params` (dict, optional): kwargs dictionary for activation function. Defaults to None.
        - `norm_layer_position` (str, optional): Position of the normalization layer relative to activation. Defaults to None. It should be 'before' or 'after' or None.
        - `norm_layer_type` (str, optional): Type of the normalization layer. Defaults to None. Examples: 'BatchNormalization', 'LayerNormalization', etc.
        - `norm_layer_params` (dict, optional): kwargs dictionary for normalization layer. Defaults to None.
        - `pool_type` (str, optional): Type of pooling layer, if any. Defaults to None. For example, 'Max', 'Avg', 'GlobalMax', 'GlobalAvg' etc.
        - `pool_kernel_size` (int, optional): Pooling kernel size. Defaults to 2. Arrays are recommended over integers.
        - `pool_padding` (int, optional): Padding for pooling layer. Defaults to 0. 'same' is NOT an option here.
        - `pool_stride` (int, optional): Pooling stride. Defaults to 1.
        - `pool_params` (dict, optional): kwargs dictionary for pooling layer module to add to or overwrite its arguments. Defaults to None.
        - `dropout` (float, optional): Dropout rate, if any. Defaults to None. for Conv?D, SpatialDropout?D is used.
        - `min_image_dim` (int, optional): Minimum image dimension. Defaults to 1. This is used for preventing the image dimension from becoming too small. 
            It can automatically adjust padding and stride for convolution and pooling layers to keep the image dimensions larger than this argument.
        - `kernel_regularizer` (regularizer, optinal): Regularizer to be used for the kernel weights in the layers.
        
        ### Returns:
        A `tf.keras.layers.Layer` instance representing a single convolutional block.
        
        ### Attributes:
        `self.output_image` (list): Size of the output image.
        `self.net` (tf.keras.models.Sequential): The actual network, a `tf.keras.models.Sequential` instance.
        """
        super(Conv_Block, self).__init__()
        self.net = tf.keras.models.Sequential()
        ret = add_conv_block(self.net, out_channels, input_shape, conv_dim, input_image, conv_kernel_size, conv_padding, conv_stride, conv_dilation, conv_params,
                             conv_activation, conv_activation_params, norm_layer_position, norm_layer_type, norm_layer_params, 
                             pool_type, pool_kernel_size, pool_padding, pool_stride, pool_params, dropout, min_image_dim, kernel_regularizer)
        self.net = ret['model']
        self._input_shape = input_shape
        self.output_image = ret['output_image']
        self._out_channels = ret['out_channels']
        self._norm_layer_position = ret['norm_layer_position']
        self._norm_layer_type = ret['norm_layer_type']
        self._min_image_dim = ret['min_image_dim']
        self._conv_padding = ret['conv_padding']
        self._conv_stride = ret['conv_stride']
        self._pool_padding = ret['pool_padding']
        self._pool_stride = ret['pool_stride']
        self._conv_module = ret['conv_module']
        self._conv_activation_module = ret['conv_activation_module']
        self._pool_module = ret['pool_module']
        self._dropout_module = ret['dropout_module']
        self._norm_layer_module = ret['norm_layer_module']
        
        # Store parameters as class attributes
        self._conv_kernel_size = conv_kernel_size
        self._conv_dilation = conv_dilation
        self._conv_params = conv_params
        self._conv_activation = conv_activation
        self._conv_activation_params = conv_activation_params
        self._norm_layer_params = norm_layer_params
        self._pool_type = pool_type
        self._pool_kernel_size = pool_kernel_size
        self._pool_params = pool_params
        self._dropout = dropout if dropout else None
        self._kernel_regularizer = kernel_regularizer

    def call(self, x, *args, **kwargs):
        return self.net(x, *args, **kwargs)
    
    def get_config(self):
        config = super(Conv_Block, self).get_config()
        hparams = {
            "input_shape":self._input_shape,
            "out_channels":self._out_channels,
            "conv_dim":self._conv_dim,
            "input_image":self._input_image,
            "conv_kernel_size":self._conv_kernel_size,
            "conv_padding":self._conv_padding,
            "conv_stride":self._conv_stride,
            "conv_dilation":self._conv_dilation,
            "conv_params":self._conv_params,
            "conv_activation":self._conv_activation,
            "conv_activation_params":self._conv_activation_params,
            "norm_layer_position":self._norm_layer_position,
            "norm_layer_type":self._norm_layer_type,
            "norm_layer_params":self._norm_layer_params,
            "pool_type":self._pool_type,
            "pool_kernel_size":self._pool_kernel_size,
            "pool_padding":self._pool_padding,
            "pool_stride":self._pool_stride,
            "pool_params":self._pool_params,
            "dropout":self._dropout,
            "min_image_dim":self._min_image_dim,
            "kernel_regularizer":self._kernel_regularizer
        }
        config['hparams'] = hparams
        return config
    
    @classmethod
    def from_config(cls, config):
        return cls(**config['hparams'])
    
    def summary(self):
        return self.net.summary()



def add_conv_block(model:tf.keras.models.Sequential, out_channels:int=None, input_shape:list=None, conv_dim:int=1, input_image:list=[30], conv_kernel_size=3, conv_padding='valid',
    conv_stride=1, conv_dilation=1, conv_params:dict=None, conv_activation:str='relu', conv_activation_params:dict=None, norm_layer_position:str=None, norm_layer_type:str=None, 
    norm_layer_params:dict=None, pool_type:str=None, pool_kernel_size=2, pool_padding:str='valid', pool_stride=1, pool_params:dict=None, dropout:float=None, min_image_dim:int=1,
    kernel_regularizer:tf.keras.regularizers.Regularizer=None):
    """ Add a Convolutional block, containing one convolution layer, followed optionally by a normalization layer, an activation layer, a pooling layer and a dropout layer,
    to the end of an existing `tf.keras.models.Sequential` instance.
    The convolution layer is mandatory, but the other ones are optional. The convolution layer can be 1D, 2D or 3D. The normalization layer can be any such layer defined
    in Keras. The activation layer can also be anything with lower-case strings, and the pooling layer can be any of the pooling layers defined for Keras. 
    The dropout layer is a spatial dropout layer. The dimension of any dropout layer will match the dimension of the convolution layer. For `Conv1D` for instance, 
    `SpatialDropout1D` will be used, if desired.

    ### Args:
    
    - `input_shape` (list|tuple): Input shape.
    - `out_channels` (int, optional): Number of convolution filters. Defaults to the input channels size.
    - `conv_dim` (int, optional): Dimension of the convolution. Defaults to 1. 1 means Conv1d, 2 means Conv2d etc.
    - `input_image` (list, optional): Size of the input image. Defaults to [30]. This must be a list/tuple of integers, with legth equal to `conv_dim`.
    - `conv_kernel_size` (int, optional): Convolution kernel size. Defaults to 3. It is strongly recommended to provide a list of integers, with length equal to `conv_dim`.
    - `conv_padding` (str, optional): Convolution padding. Defaults to 'same'. Arrays are recommended over integers.
    - `conv_stride` (int, optional): Convolution stride. Defaults to 1.
    - `conv_dilation` (int, optional): Convolution dilation. Defaults to 1.
    - `conv_params` (dict, optional): Additional dictionary of kwargs for constructor of Conv?d module. Defaults to None.
      If provided, this dictionary will not only add to, but also overwrite any existing arguments passed to the Conv?D constructor.
    - `conv_activation` (str, optional): String representing activation function. Defaults to 'relu'. Examples: 'leakyrelu', 'sigmoid', 'tanh' etc.
    - `conv_activation_params` (dict, optional): kwargs dictionary for activation function. Defaults to None.
    - `norm_layer_position` (str, optional): Position of the normalization layer relative to activation. Defaults to None. It should be 'before' or 'after' or None.
    - `norm_layer_type` (str, optional): Type of the normalization layer. Defaults to None. Examples: 'BatchNormalization', 'LayerNormalization', etc.
    - `norm_layer_params` (dict, optional): kwargs dictionary for normalization layer. Defaults to None.
    - `pool_type` (str, optional): Type of pooling layer, if any. Defaults to None. For example, 'Max', 'Avg', 'GlobalMax', 'GlobalAvg' etc.
    - `pool_kernel_size` (int, optional): Pooling kernel size. Defaults to 2. Arrays are recommended over integers.
    - `pool_padding` (int, optional): Padding for pooling layer. Defaults to 0. 'same' is NOT an option here.
    - `pool_stride` (int, optional): Pooling stride. Defaults to 1.
    - `pool_params` (dict, optional): kwargs dictionary for pooling layer module, to add to or overwrite existing parameters. Defaults to None.
    - `dropout` (float, optional): Dropout rate, if any. Defaults to None. for Conv?D, SpatialDropout?D is used.
    - `min_image_dim` (int, optional): Minimum image dimension. Defaults to 1. This is used for preventing the image dimension from becoming too small. 
        It can automatically adjust padding and stride for convolution and pooling layers to keep the image dimensions larger than this argument.
    - `kernel_regularizer` (regularizer, optional): Regularizer to be used for the kernel weights in the layers.
    
    ### Returns:
    A dictionary with the following fields:
    ```python
    {
        'model'
        'output_image'
        'out_channels'
        'norm_layer_position'
        'norm_layer_type'
        'min_image_dim'
        'conv_padding'
        'conv_stride'
        'pool_padding'
        'pool_stride'
        'conv_module'
        'conv_activation_module'
        'pool_module'
        'dropout_module'
        'norm_layer_module'
    }
    ```
    """
    # Output channels check
    if not (isinstance(out_channels,int) and out_channels > 0):
        warnings.warn("Invalid value out_channels={}. Using value {} equal to input_shape[-1].".format(out_channels,input_shape), UserWarning)
        out_channels = input_shape[-1]
    # Convolution dimension check    
    assert isinstance(conv_dim, int) and conv_dim in [1,2,3], "`conv_dim` must be an integer among [1,2,3], not {} which has type {}.".format(conv_dim, str(type(conv_dim)))
    # Determine convolution module used
    _conv_module = convdict_keras["conv{}d".format(conv_dim)]
    # Input image size check
    assert isinstance(input_image, (list,tuple)) and len(input_image)==conv_dim, \
        "`input_image` must be a list or tuple of length equal to `conv_dim`, not {} which has type {}.".format(input_image, str(type(input_image)))
    # Check activation function and module
    _conv_activation_module = actdict_keras[conv_activation]  if conv_activation else None
    # Check position, type and parameters of normalization layer
    if not norm_layer_position in ['before', 'after', None]:
        warnings.warn(("Invalid value {} for `norm_layer_position`: It can only be 'before' (before activation), 'after' (after activation) or None. "+
                        "Using default value of None. There will be no normalization.").format(norm_layer_position), UserWarning)
        norm_layer_position = None
    if norm_layer_position is None: norm_layer_type = None
    if norm_layer_type is None: norm_layer_position = None
    _norm_layer_module = getattr(tf.keras.layers, norm_layer_type) if norm_layer_type else None
    # Check pooling layer type, module and parameters
    _pool_module = getattr(tf.keras.layers, "{}Pool{}D".format(pool_type, conv_dim)) if pool_type else None
    # Check Dropout parameters
    _dropout_module = getattr(tf.keras.layers, 'SpatialDropout{}D'.format(conv_dim)) if dropout else None
    # Store minimum desired image size
    min_image_dim = min_image_dim if min_image_dim>0 else 1
    img_size = input_image
    # -----------------------------------------------------------------------------        
    # Check if output image size is smaller than min_image_dim, and adjust parameters if necessary
    # print("Running calc_size for input image size {}".format(img_size))
    temp_img_size = calc_image_size(img_size, kernel_size=conv_kernel_size, stride=conv_stride, padding=conv_padding, dilation=conv_dilation)
    if min(temp_img_size) < min_image_dim:
        warnings.warn(
            "Output image ({}) of the convolution operation is smaller in one or more dimensions than min_image_dim={} for add_conv_block(). ".format(temp_img_size,min_image_dim)+ 
            "Using padding='same' and stride=1 instead of padding={} and stride={}".format(conv_padding, conv_stride), UserWarning)
        conv_padding = 'same'
        conv_stride = 1
    # Construct convolutional layer
    _kwargs = {'filters':out_channels, 'kernel_size':conv_kernel_size, 'strides':conv_stride, 'padding':conv_padding, 'dilation_rate':conv_dilation, 
               'kernel_regularizer':kernel_regularizer}
    if input_shape: _kwargs.update({'input_shape':input_shape})
    if conv_params: _kwargs.update(conv_params)
    model.add(_conv_module(**_kwargs))
    # Calculate output image size
    img_size = calc_image_size(img_size, kernel_size=conv_kernel_size, stride=conv_stride, padding=conv_padding, dilation=conv_dilation)
    # ---------------------------------------------------------------------------
    # Construct normalization layer, if it should be here.
    if norm_layer_position=='before':
        if norm_layer_params: model.add(_norm_layer_module(**norm_layer_params))
        else: model.add(_norm_layer_module())
    # Construct activation layer
    if conv_activation:
        if conv_activation_params: model.add(tf.keras.layers.Activation(_conv_activation_module(**conv_activation_params)))
        else: model.add(tf.keras.layers.Activation(_conv_activation_module))
    # Construct normalization layer, if it should be here.
    if norm_layer_position=='after':
        if norm_layer_params: model.add(_norm_layer_module(**norm_layer_params))
        else: model.add(_norm_layer_module())
    # ---------------------------------------------------------------------------
    # Check type and parameters of the pooling layer, and calculate output image size
    if pool_type is not None and 'global' in pool_type.lower():
        if pool_params: model.add(_pool_module(**pool_params))
        else: model.add(_pool_module())
        img_size = [1]*conv_dim
    elif pool_type is not None:
        temp_img_size = calc_image_size(img_size, kernel_size=pool_kernel_size, stride=pool_stride, padding=pool_padding, dilation=1) # Keras pool has no support for dilation :(
        if min(temp_img_size) < min_image_dim:
            warnings.warn(
                "Output image ({}) of the pooling operation is smaller in one or more dimensions than min_image_dim={} for add_conv_block(). ".format(temp_img_size,min_image_dim)+ 
                "Using padding={} and stride=1 instead of padding={} and stride={}".format('same', pool_padding, pool_stride), UserWarning)
            pool_padding = 'same'
            pool_stride = 1
        _kwargs = {'pool_size':pool_kernel_size, 'strides':pool_stride, 'padding':pool_padding}
        if pool_params: _kwargs.update(pool_params)
        model.add(_pool_module(**_kwargs))
        img_size = calc_image_size(img_size, kernel_size=pool_kernel_size, stride=pool_stride, padding=pool_padding, dilation=1)
    # ---------------------------------------------------------------------------
    # Construct Dropout layer    
    if dropout: model.add(_dropout_module(dropout))
    # Store output image size as attribute    
    output_image = img_size
    d = {
        'model':model, 'output_image':output_image, 'out_channels':out_channels, 'norm_layer_position':norm_layer_position,
        'norm_layer_type':norm_layer_type, 'min_image_dim':min_image_dim, 'conv_padding':conv_padding, 'conv_stride':conv_stride, 'pool_padding':pool_padding,
        'pool_stride':pool_stride, 'conv_module':_conv_module, 'conv_activation_module':_conv_activation_module, 'pool_module':_pool_module,
        'dropout_module':_dropout_module, 'norm_layer_module':_norm_layer_module
    }
    return d




if __package__=="eznet_keras.models":
    from .keras_smart_module import *
    from .dense_block import *
else:
    import sys, os
    sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))
    from keras_smart_module import *
    from dense_block import *

class ANN(KerasSmartModel):
    
    sample_hparams = {
        "model_name": "ANN",
        "input_size": 10,
        "output_size": 3,
        "width": 32,
        "depth": 2,
        "hidden_activation": "relu",
        "hidden_activation_params": None,
        "output_activation": None,
        "output_activation_params": None,
        "norm_layer_type":"BatchNormalization",
        "norm_layer_position": "before",
        "norm_layer_params": None,
        "dropout": 0.2,
        "learning_rate": 0.001,
        "learning_rate_decay_gamma": 0.99,
        "batch_size": 32,
        "epochs": 2,
        "validation_tolerance_epochs": 2,
        "validation_data":[0.05,'trainset'],
        "l2_reg": 0.0001,
        "loss_function": "categorical_crossentropy",
        "loss_function_params": None,
        'optimizer': 'Adam',
        'optimizer_params': None,
        'metrics':['accuracy'],
        'checkpoint_path':None,
        'early_stopping_monitor':'loss',
        'early_stopping_mode':'min',
        'early_stopping_value':1.0e-6
    }
    
    
    def __init__(self, hparams:dict=None):
        """Typical Artificial Neural Network class, also known as multilayer perceptron. This class will create a fully connected feedforward artificial neural network.
        It can be used for classification, regression, etc. It basically encompasses enough options to build all kinds of ANNs with any number of 
        inputs, outputs, layers with custom or arbitrary width or depth, etc. Supports multiple activation functions for hidden layers and the output layer.
        
        ### Usage
        `net = ANN(hparams)` where `hparams` is the dictionary of hyperparameters.

        It can include the following keys:
            - `input_size` (int): number of inputs to the ANN, i.e. size of the input layer.
            - `output_size` (int): number of outputs to predict, i.e. size of the output layer.
            - `width` (int|list): (list of) hidden layer widths. 
                a number sets them all the same, and a list/array sets each hidden layer according to the list.
            - `depth` (int): Specifies the depth of the network (number of hidden layers).
                It must be specified unless `width` is provided as a list. Then the depth will be inferred form it.
            - `hidden_activation` (str): (list of) Activations of the hidden layers. Examples include "relu", "leakyrelu", "sigmoid", "tanh", "softmax", etc.
            - `hidden_activation_params` (dict): (list of) Parameters for the hidden layer activation function, if any.
            - `output_activation` (str): Activation of the output layer, if any.
                **Note**: For classification problems, you may want to choose "sigmoid" or "softmax".
                That being said, you usually don't need to specify an activation for the output layer at all, if e.g. 'from_logits' is used.
                **Note**: For regression problems, no activation is needed. It is by default linear, unless you want to manually specify an activation.
            - `output_activation_params` (dict): Parameters for the output activation function, if any.
            - `norm_layer_type` (str): (list of) Types of normalization layers to use for each hidden layer. Options are "BatchNormalization", "LayerNormalization", etc.
            - `norm_layer_position` (str): (list of) where the normalization layer should be included relative to the activation function, 'before' or 'after'.
            - `norm_layer_params` (dict): (list of) Dictionaries of parameters for the normalization layers.
            - `dropout` (float): (list of) the dropout rates after every hidden layer. It should be a probability value between 0 and 1.
            - `learning_rate` (float): Initial learning rate of training.
            - `learning_rate_decay_gamma` (float): Exponential decay rate gamma for learning rate, if any.
            - `optimizer` (str): Optimizer. Examples: "Adam", "SGD" ,"RMSProp", etc.
            - `optimizer_params` (dict): Additional parameters of the optimizer, if any.
            - `batch_size` (int): Minibatch size for training.
            - `epochs` (int): Maximum number of epochs for training.
            - `validation_tolerance_epochs` (int): Epochs to tolerate unimproved val loss, before early stopping.
            - `validation_data` (list): List of [validation_split, 'trainset'|'testset'].
            - `l2_reg` (float): L2 regularization parameter.
            - `loss_function` (str): Loss function. Examples: "mse","binary_crossentropy", "categorical_crossentropy", etc.
            - `loss_function_params` (dict): Additional parameters for the loss function, if any.
            - `metrics` (list): list of metrics for Keras compilation, e.g. ['accuracy'].
            - `checkpoint_path` (str): Path to the directory where checkpoints will be saved at every epoch.
            - `early_stopping_monitor` (str): Monitor whose critical value will cause early stopping. Default is 'loss', but 'val_loss' is typically used.
            - `early_stopping_mode` (str): Mode of the parameter whose critical value will be used for early stopping. Deafults to 'min' for any error. 'max' is for accuracy, etc.
            - `early_stopping_value` (float): Value of the monitor at which point training will stop becasue the critical value has been reached.
        
        Note that for all such hyperparameters that have a (list of) at the beginning, the entry can be a single item repeated for all hidden layers, or it can be a list of items
        for all hidden layers. If a list is provided, it must have the same length as the depth of the network. Also note that depth does not include the input and output layers.
        This gives you the ability to specify different width, dropout rate, normalization layer and its parameters, and so forth.
        
        Also note that the hidden layer just before the output layer will not have any dropout, which is typical.

        ### Returns
        It returns a `tf.keras.models.Model` object that corresponds with an ANN model.
        run `net.summary()` afterwards to see what the ANN holds.
        The returned module is a `KerasSmartModel` object, which is a subclass. It has built-in functions for training, evaluation, etc.
        """
        super(ANN, self).__init__(hparams)
        # Read and store hyperparameters
        if not hparams: hparams=self.sample_hparams
        self._insize = hparams["input_size"]
        self._outsize = hparams["output_size"]
        self._dropout = hparams.get("dropout")
        self._width = hparams.get("width")
        self._depth = hparams.get("depth")
        self._denseactivation = hparams["hidden_activation"]
        self._denseactivation_params = hparams.get("hidden_activation_params")
        self._outactivation = hparams.get("output_activation") if hparams.get("output_activation") else None
        self._outactivation_params = hparams.get("output_activation_params")
        self._norm_layer_type = hparams.get("norm_layer_type")
        self._norm_layer_position = hparams.get("norm_layer_position")
        self._norm_layer_params = hparams.get("norm_layer_params")
        self.batch_input_shape = (self._batch_size, self._insize)
        self.batch_output_shape = (self._batch_size, self._outsize)
        self._out_activation_module = actdict_keras[self._outactivation] if self._outactivation else None
        
        # Generate arrays containing parameters of each Dense Block (Every block contains a linear, normalization, activation, and dropout layer).
        self._dense_width_vec = self._gen_hparam_vec_for_dense(self._width, 'width')
        self._dense_activation_vec = self._gen_hparam_vec_for_dense(self._denseactivation, 'hidden_activation')
        self._dense_activation_params_vec = self._gen_hparam_vec_for_dense(self._denseactivation_params, 'hidden_activation_params')
        self._dense_norm_layer_type_vec = self._gen_hparam_vec_for_dense(self._norm_layer_type, 'norm_layer_type')
        self._dense_norm_layer_params_vec = self._gen_hparam_vec_for_dense(self._norm_layer_params, 'norm_layer_params')
        self._dense_norm_layer_position_vec = self._gen_hparam_vec_for_dense(self._norm_layer_position, 'norm_layer_position')
        self._dense_dropout_vec = self._gen_hparam_vec_for_dense(self._dropout, 'dropout')
        
        # Initialize sequential model
        self.net = tf.keras.models.Sequential()
        
        # Construct the dense layers
        in_size = self._insize
        for i in range(self._depth):
            out_size = self._dense_width_vec[i]
            temp_dropout_rate = self._dense_dropout_vec[i] if (i != self._depth-1) else None # The hidden layer just before the output layer rarely has Dropout.
            _kwargs = {
                'model':self.net,
                'output_size':self._dense_width_vec[i],
                'activation':self._dense_activation_vec[i],
                'activation_params':self._dense_activation_params_vec[i],
                'norm_layer_type':self._dense_norm_layer_type_vec[i],
                'norm_layer_position':self._dense_norm_layer_position_vec[i],
                'norm_layer_params':self._dense_norm_layer_params_vec[i],
                'dropout':temp_dropout_rate,
                'kernel_regularizer':(tf.keras.regularizers.L2(self._l2_reg) if self._l2_reg else None)
            }
            if i==0:
                _kwargs.update({'input_shape':[self._insize]})
            add_dense_block(**_kwargs)
            in_size = out_size
        # Output layer
        self.net.add(tf.keras.layers.Dense(self._outsize))
        if self._outactivation:
            if self._outactivation_params: self.net.add(tf.keras.layers.Activation(self._out_activation_module(**self._outactivation_params)))
            else: self.net.add(tf.keras.layers.Activation(self._out_activation_module))

    def _gen_hparam_vec_for_dense(self, hparam, hparam_name, **kwargs):
        return generate_array_for_hparam(hparam, self._depth, hparam_name=hparam_name, count_if_not_list_name='depth', **kwargs)
        
    def call(self, x, **kwargs):
        return self.net(x, **kwargs)




if __package__=="eznet_keras.models":
    from ..utils import *
else:
    import os, sys
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    sys.path.append(parent_dir)
    from utils import *


class Dense_Block(tf.keras.layers.Layer):
    def __init__(self, input_shape:list, output_size:int=None, activation:str=None, activation_params:dict=None, norm_layer_type:str=None, norm_layer_position:str='before', 
                 norm_layer_params:dict=None, dropout:float=None, kernel_regularizer:tf.keras.regularizers.Regularizer=None):
        """Dense (fully connected) block containing one linear layer, followed optionally by a normalization layer, an activation function and a Dropout layer.

        ### Args:
            - `input_shape` (list|tuple): Shape of the input.
            - `output_size` (int, optional): Number of output features. Defaults to None, in which case it will be input_size.
            - `activation` (str, optional): Activation function in string form. Defaults to None. Examples: 'relu', 'leakyrelu', 'tanh', 'sigmoid', etc.
            - `activation_params` (dict, optional): kwargs to pass to the activation function constructor. Defaults to None.
            - `norm_layer_type` (str, optional): Type of normalization layer. Defaults to None. Examples: 'BatchNormalization', 'LayerNormalization', etc.
            - `norm_layer_position` (str, optional): Position of norm layer relative to activation. Defaults to 'before'. Alternative is 'after'.
            - `norm_layer_params` (dict, optional): kwargs to pass to the norm layer constructor. Defaults to None.
            - `dropout` (float, optional): Dropout rate at the end. Defaults to None. Must be a float between 0 and 1.
            - `kernel_regularizer` (regularizer, optinal): Regularizer to be used for the kernel weights in the layers.
            
        ### Returns:
        A `tf.keras.layers.Layer` object.
        """
        super(Dense_Block, self).__init__()
        if output_size is None: output_size = input_shape[-1]
        self._input_shape = input_shape
        self._output_size = output_size
        self._activation = activation
        self._activation_params = activation_params
        self._norm_layer_type = norm_layer_type
        self._norm_layer_position = norm_layer_position
        self._norm_layer_params = norm_layer_params
        self._dropout = dropout
        self._activation_module = actdict_keras[activation] if activation else None
        self._norm_layer_module = getattr(tf.keras.layers, norm_layer_type) if norm_layer_type else None
        self._dropout_module = tf.keras.layers.Dropout if dropout else None
        self._kernel_regularizer = kernel_regularizer
        self.net = tf.keras.models.Sequential()
        self.net.add(tf.keras.layers.Dense(output_size, input_shape=input_shape, kernel_regularizer=kernel_regularizer))
        if norm_layer_type and norm_layer_position=='before': 
            if norm_layer_params: self.net.add(self._norm_layer_module(**norm_layer_params))
            else: self.net.add(self._norm_layer_module())
        if activation: 
            if activation_params: self.net.add(tf.keras.layers.Activation(self._activation_module(**activation_params)))
            else: self.net.add(tf.keras.layers.Activation(self._activation_module))
        if norm_layer_type and norm_layer_position=='after': 
            if norm_layer_params: self.net.add(self._norm_layer_module(**norm_layer_params))
            else: self.net.add(self._norm_layer_module())
        if dropout: self.net.add(self._dropout_module(dropout))
    
    def call(self, x, *args, **kwargs):
        return self.net(x, *args, **kwargs)
    
    def get_config(self):
        config = super(Dense_Block, self).get_config()
        hparams = {
            'input_shape':self._input_shape,
            'output_size':self._output_size,
            'activation':self._activation,
            'activation_params':self._activation_params,
            'norm_layer_type':self._norm_layer_type,
            'norm_layer_position':self._norm_layer_position,
            'norm_layer_params':self._norm_layer_params,
            'dropout':self._dropout,
            'kernel_regularizer':self._kernel_regularizer
        }
        config['hparams'] = hparams
        return config
    
    @classmethod
    def from_config(cls, config):
        return cls(**config['hparams'])
    
    def summary(self):
        return self.net.summary()
        



def add_dense_block(model:tf.keras.models.Sequential, output_size:int, input_shape:list=None, activation:str=None, activation_params:dict=None, 
                    norm_layer_type:str=None, norm_layer_position:str='before', norm_layer_params:dict=None, dropout:float=None, 
                    kernel_regularizer:tf.keras.regularizers.Regularizer=None):
    """Add a Dense (fully connected) block containing one linear layer, followed optionally by a normalization layer, an activation function and a Dropout layer, 
    to a `tf.keras.models.Sequential` instance.

        ### Args:
            - `input_shape` (list|tuple, optional): Shape of the input.
            - `output_size` (int, optional): Number of output features. Defaults to None, in which case it will be input_size.
            - `activation` (str, optional): Activation function in string form. Defaults to None. Examples: 'relu', 'leakyrelu', 'tanh', 'sigmoid', etc.
            - `activation_params` (dict, optional): kwargs to pass to the activation function constructor. Defaults to None.
            - `norm_layer_type` (str, optional): Type of normalization layer. Defaults to None. Examples: 'BatchNormalization', 'LayerNormalization', etc.
            - `norm_layer_position` (str, optional): Position of norm layer relative to activation. Defaults to 'before'. Alternative is 'after'.
            - `norm_layer_params` (dict, optional): kwargs to pass to the norm layer constructor. Defaults to None.
            - `dropout` (float, optional): Dropout rate at the end. Defaults to None. Must be a float between 0 and 1.
            - `kernel_regularizer` (regularizer, optinal): Regularizer to be used for the kernel weights in the layers.
            
        ### Returns:
        Nothing. It modifies the `model` argument passed to it.
        """
    _activation_module = actdict_keras[activation] if activation else None
    _norm_layer_module = getattr(tf.keras.layers, norm_layer_type) if norm_layer_type else None
    _dropout_module = tf.keras.layers.Dropout if dropout else None
    if input_shape:
        model.add(tf.keras.layers.Dense(output_size, input_shape=input_shape, kernel_regularizer=kernel_regularizer))
    else:
        model.add(tf.keras.layers.Dense(output_size, kernel_regularizer=kernel_regularizer))
    if norm_layer_type and norm_layer_position=='before': 
        if norm_layer_params: model.add(_norm_layer_module(**norm_layer_params))
        else: model.add(_norm_layer_module())
    if activation: 
        if activation_params: model.add(tf.keras.layers.Activation(_activation_module(**activation_params)))
        else: model.add(tf.keras.layers.Activation(_activation_module))
    if norm_layer_type and norm_layer_position=='after': 
        if norm_layer_params: model.add(_norm_layer_module(**norm_layer_params))
        else: model.add(_norm_layer_module())
    if dropout: model.add(_dropout_module(dropout))


if __package__=="eznet_keras.models":
    from .keras_smart_module import *
    from .conv_block import *
    from .dense_block import *
else:
    import sys, os
    sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))
    from keras_smart_module import *
    from conv_block import *
    from dense_block import *

class Conv_Network(KerasSmartModel):
    sample_hparams = {
        "model_name": "Conv_Network",
        # I/O shapes (without the batch dimension)
        "input_shape": [28, 28, 3],
        "output_shape": [10],
        # Convolution blocks
        "num_conv_blocks": 2,
        "conv_dim": 2,
        "conv_params": None,
        "conv_channels": "auto",
        "conv_kernel_size": 3,
        "conv_padding": "valid",
        "conv_stride": 1,
        "conv_dilation": 1,
        "conv_activation": "relu",
        "conv_activation_params": None,
        "conv_norm_layer_type": "BatchNormalization",
        "conv_norm_layer_position": "before",
        "conv_norm_layer_params": None,
        "conv_dropout": 0.1,
        "pool_type": "Max",
        "pool_kernel_size": 2,
        "pool_padding": 'valid',
        "pool_stride": 1,
        "pool_params": None,
        "min_image_size": 4,
        # Fully connected blocks
        "dense_width": "auto",
        "dense_depth": 2,
        "dense_activation": "relu",
        "dense_activation_params": None,
        "output_activation": "softmax",
        "output_activation_params": None,
        "dense_norm_layer_type": "BatchNormalization",
        "dense_norm_layer_position": "before",
        "dense_norm_layer_params": None,
        "dense_dropout": 0.1,
        # Training procedure
        "l2_reg": 0.0001,
        "batch_size": 32,
        "epochs": 2,
        "validation_data": [0.05,'testset'],
        "validation_tolerance_epochs": 5,
        "learning_rate": 0.01,
        "learning_rate_decay_gamma": 0.9,
        "loss_function": "categorical_crossentropy",
        "optimizer": "Adam",
        "optimizer_params": None,
        'metrics':['accuracy'],
        'checkpoint_path':None,
        'early_stopping_monitor':'loss',
        'early_stopping_mode':'min',
        'early_stopping_value':1.0e-6
    }
    
    
    def __init__(self, hparams:dict=None):
        """Standard Convolutional Neural Network, containing convolutional blocks followed by fully-connected blocks. It supports 1D, 2D, and 3D convolutions, and can be used for 
        image classification, timeseries classification, video classification, and so forth. The module can easily be trained and evaluated using its own methods,
        because it inherits from `KerasSmartModel`.

        ### Usage

        `model = Conv_Network(hparams)` where `hparams` is dictionary of hyperparameters containing the following:

        #### I/O shapes
        
        - `input_shape` (list): Input shape *WITHOUT* the batch dimension. For instance, for 2D images, input should be [N, H, W, C], therefore `input_shape` should be [H, W, C].
        - `output_shape` (int): Output shape *WITHOUT* the batch dimension. For instance, for K-class classification, model outputs can be [N, K], so `output_shape` should be [K].
            
        #### Convolution blocks
        
        - `num_conv_blocks` (int): Number of convolutional blocks. Every block contains a convolutional layer, and
            optionally a normalization layer, an activation layer, a pooling layer, and finally a dropout layer.
        - `conv_dim` (int): Dimensionality of the convolution. 1, 2, or 3.
        - `conv_params` (dict): (list of) kwargs dict to pass to the convolution constructor in each block. Defaults to None.
        - `conv_channels` (int|list|str): (list of) Number of filters of the convolution layer in each conv block. If `auto`, it will start
            with the input channels, and double with every block, in powers of two. If `list`, it should be a list
            of channels for each conv block. If `int`, it will be the same for all conv blocks. Default is `auto`.
        - `conv_kernel_size` (int|list): (list of) Kernel size of the convolution layers. Should be a list of integers,
            a list of tuples of integers (for conv2d or conv3d), or an integer. If it is a list, it MUST have the same 
            length as `num_conv_blocks`. If it is an integer, it will be the same for all conv blocks. Defaults to 3.
        - `conv_padding` (int|str|list): (list of) Paddings of convolution layers. Format is as `conv_kernel_size`. Defaults to "valid".
        - `conv_stride` (int|list): (list of) Stride of convolution layers. Format is as `conv_kernel_size`. Defaults to 1.
        - `conv_dilation` (int|list): (list of) Dilation of convolution layers. Format is as `conv_kernel_size`. Defaults to 1.
        - `conv_activation` (str|list): (list of) string(s) representing activation func of the convolution layers. Examples: 'relu', 'leakyrelu', 'sigmoid', 'tanh', etc.
        - `conv_activation_params` (dict|list): (list of) dicts for the convolution activation functions' constructors. Defaults to None.
        - `conv_norm_layer_type` (str|list): (list of) types of normalization layers to use in the conv blocks. Examples: 'BatchNormalization', 'LayerNormalization', etc.
            Defaults to None.
        - `conv_norm_layer_position` ("before"|"after"|list): (list of) positions of the normalization layers in the 
            convolutional blocks relative to the activation functions. Defaults to "before". If it is a list, it should be a list of strings of the same length as `num_conv_blocks`
        - `conv_norm_layer_params` (dict|list): (list of) kwargs dict for the convolution normalization layers' constructors. Defaults to None.    
        - `conv_dropout` (float|list): (list of) Dropout rates of the convolution blocks. Defaults to None.
        - `pool_type` (str|list): (list of) types of pooling layer. "Max", "Avg", "GlobalMax", "GlobalAvg", etc. Defaults to None, in which case there will be no pooling layer.
        - `pool_kernel_size` (int|list): (list of) kernel sizes of the pooling layers, with similar format to 
            `conv_kernel_size`. Again, it can be a list of integers, a list of tuples of integers, or an integer.
        - `pool_padding` (str|list): (list of) paddings of the pooling layers.
        - `pool_stride` (int|list): (list of) strides of the pooling layers.
        - `pool_params` (dict|list): (list of) kwargs dicts for the pooling layers' constructors.
        - `min_image_size` (int): Minimum size of the image to be reduced to in convolutions and poolings.
            After this point, the padding and striding will be chosen such that image size does not decrease further. Defaults to 1.
            
        #### Dense blocks
        
        - `dense_width` ("auto"|int|list): Width of the hidden layers of the Dense network. "auto", a number (for all of them) or a list holding width of each hidden layer.
            If "auto", it will start with the output size of the Flatten() layer, halving at every Dense block.
        - `dense_depth` (int): Depth (number of hidden layers) of the Dense network.
        - `dense_activation` (str|list): (list of) activation function for hidden layers of the Dense network. Examples: 'relu', 'leakyrelu', 'sigmoid', 'tanh', etc.
        - `dense_activation_params` (dict|list): (list of) dicts for the dense activation functions' constructors.
        - `output_activation` (str): Activation function for the output layer of the Dense network, if any.
            **NOTE** Depending on the loss function, you may not need an activation function.
        - `output_activation_params` (dict): Dictionary of parameters for the output activation function's constructor.
        - `dense_norm_layer_type` (str|list): (list of) types of normalization layers to use in the dense blocks. Examples: 'BatchNormalization', 'LayerNormalization', etc.
            Defaults to None, in which case no normalization layer will be used.
        - `dense_norm_layer_position` ("before"|"after"|list): (list of) positions of the normalization layers in the dense blocks relative to the activation functions. 
            Defaults to "before". If it is a list, it should be a list of strings of the same length as `dense_depth`.
        - `dense_norm_layer_params` (dict|list): (list of) kwargs dict for the dense normalization layers' constructors.
        - `dense_dropout` (float|list): (list of) Dropout rates (if any) for the hidden layers of the Dense network.
        
        #### Training procedure
        
        - `batch_size` (int): Minibatch size, the expected input size of the network.
        - `learning_rate` (float): Initial learning rate of training.
        - `learning_rate_decay_gamma` (float): Exponential decay rate gamma for learning rate, if any.
        - `optimizer` (str): Optimizer. Examples: 'Adam', 'SGD', 'RMSprop', etc.
        - `optimizer_params` (dict): Additional parameters of the optimizer, if any.
        - `epochs` (int): Maximum number of epochs for training.
        - `validation_tolerance_epochs` (int): Epochs to tolerate unimproved val loss, before early stopping.
        - `l2_reg` (float): L2 regularization parameter.
        - `loss_function` (str): Loss function. Examples: 'mse','categorical_crossentropy',etc.
        - `loss_function_params` (dict): Additional parameters for the loss function, if any.
        - `validation_data` (tuple): Validation data, if any. It should be a tuple of (portion, from_dataset). For instance, [0.05, 'testset'] means 5% of the testset will be used 
            for validation.The second element of the tuple can only be 'trainset' and 'testset'. The first element must be a float between 0 and 1. 
            If the second element is not specified, testset will be used by default.
        - `metrics` (list): list of metrics for Keras compilation, e.g. ['accuracy'].
        - `checkpoint_path` (str): Path to the directory where checkpoints will be saved at every epoch.
        - `early_stopping_monitor` (str): Monitor whose critical value will cause early stopping. Default is 'loss', but 'val_loss' is typically used.
        - `early_stopping_mode` (str): Mode of the parameter whose critical value will be used for early stopping. Deafults to 'min' for any error. 'max' is for accuracy, etc.
        - `early_stopping_value` (float): Value of the monitor at which point training will stop becasue the critical value has been reached.
        
        ### Returns
        
        - Returns a `tf.keras.models.Model` object that can be trained and used accordingly.
        - Run `net.summary()` afterwards to see what you have inside the network.
        - A `KerasSmartModel` object is returned. This module has its own functions for training, evaluation, etc.
        """
        super(Conv_Network, self).__init__(hparams)
        if not hparams: hparams = self.sample_hparams
        # Input and output shapes
        self.model_name = hparams["model_name"] if hparams.get("model_name") else "Conv_Network"
        self._input_shape = hparams["input_shape"]
        self._output_shape = hparams["output_shape"]
        self._N = int(hparams["batch_size"])
        self.batch_input_shape = list(self._input_shape).copy()
        self.batch_input_shape.insert(0, self._N)
        self.batch_output_shape = list(self._output_shape).copy()
        self.batch_output_shape.insert(0, self._N)
        self.size_list = [self._input_shape]
    
        # Initialiaing sequential network
        self.net = tf.keras.models.Sequential()
        
        # Convolutional layers hyperparameters
        self._num_conv_blocks = hparams.get("num_conv_blocks")
        self._conv_dim = hparams.get("conv_dim")
        self._conv_params = hparams.get("conv_params")    
        self._conv_channels = hparams.get("conv_channels") if hparams.get("conv_channels") else "auto"
        self._conv_kernel_size = hparams.get("conv_kernel_size") if hparams.get("conv_kernel_size") else 3
        self._conv_padding = hparams["conv_padding"] if hparams.get("conv_padding") else "valid"
        self._conv_stride = hparams["conv_stride"] if hparams.get("conv_stride") else 1
        self._conv_dilation = hparams["conv_dilation"] if hparams.get("conv_dilation") else 1
        self._conv_activation = hparams["conv_activation"] if hparams.get("conv_activation") else "relu"
        self._conv_activation_params = hparams.get("conv_activation_params")
        self._conv_norm_layer_type = hparams.get("conv_norm_layer_type")
        self._conv_norm_layer_position = hparams.get("conv_norm_layer_position")
        self._conv_norm_layer_params = hparams.get("conv_norm_layer_params")
        self._conv_dropout = hparams.get("conv_dropout")
        self._pool_type = hparams.get("pool_type")
        self._pool_kernel_size = hparams.get("pool_kernel_size") if hparams.get("pool_kernel_size") else 2
        self._pool_padding = hparams["pool_padding"] if hparams.get("pool_padding") else 'valid'
        self._pool_stride = hparams["pool_stride"] if hparams.get("pool_stride") else 1
        self._pool_params = hparams.get("pool_params")
        self._min_image_size = hparams["min_image_size"] if hparams.get("min_image_size") else 1        
        
        # Generate lists of hyperparameters for conv/pool layers
        self._conv_channels_vec = self._gen_hparam_vec_for_conv(self._conv_channels, "conv_channels", 
            check_auto=True, init_for_auto=self._input_shape[-1], powers_of_two_if_auto=True, direction_if_auto="up")
        self._conv_kernel_size_vec = self._gen_hparam_vec_for_conv(self._conv_kernel_size, "conv_kernel_size")
        self._pool_kernel_size_vec = self._gen_hparam_vec_for_conv(self._pool_kernel_size, 'pool_kernel_size')
        self._conv_padding_vec = self._gen_hparam_vec_for_conv(self._conv_padding, 'conv_padding')
        self._pool_padding_vec = self._gen_hparam_vec_for_conv(self._pool_padding, 'pool_padding')
        self._conv_stride_vec = self._gen_hparam_vec_for_conv(self._conv_stride, 'conv_stride')
        self._pool_stride_vec = self._gen_hparam_vec_for_conv(self._pool_stride, 'pool_stride')
        self._conv_dilation_vec = self._gen_hparam_vec_for_conv(self._conv_dilation, 'conv_dilation')
        self._conv_activation_vec = self._gen_hparam_vec_for_conv(self._conv_activation, 'conv_activation')
        self._conv_activation_params_vec = self._gen_hparam_vec_for_conv(self._conv_activation_params, 'conv_activation_params')
        self._pool_type_vec = self._gen_hparam_vec_for_conv(self._pool_type, 'pool_type')
        self._pool_params_vec = self._gen_hparam_vec_for_conv(self._pool_params, 'pool_params')
        self._conv_params_vec = self._gen_hparam_vec_for_conv(self._conv_params, 'conv_params')
        self._conv_norm_layer_type_vec = self._gen_hparam_vec_for_conv(self._conv_norm_layer_type, 'conv_norm_layer_type')
        self._conv_norm_layer_params_vec = self._gen_hparam_vec_for_conv(self._conv_norm_layer_params, 'conv_norm_layer_params')
        self._conv_norm_layer_position_vec = self._gen_hparam_vec_for_conv(self._conv_norm_layer_position, 'conv_norm_layer_position')
        self._conv_dropout_vec = self._gen_hparam_vec_for_conv(self._conv_dropout, 'conv_dropout')
        
        # Constructing the encoder (convolutional blocks)
        # print("input_shape: ", self._input_shape)
        in_channels = self._input_shape[-1]
        input_image = list(self._input_shape[:-1])
        for i in range(self._num_conv_blocks):
            out_channels = self._conv_channels_vec[i]
            # print("in_channels: ", in_channels)
            # print("out_channels: ", out_channels)
            # print("input_image: ", input_image)
            _kwargs = {
                'model':self.net, 
                'out_channels':out_channels, 
                'conv_dim':self._conv_dim, 
                'input_image':input_image, 
                'conv_kernel_size':self._conv_kernel_size_vec[i], 
                'conv_padding':self._conv_padding_vec[i],
                'conv_stride':self._conv_stride_vec[i], 
                'conv_dilation':self._conv_dilation_vec[i], 
                'conv_params':self._conv_params_vec[i], 
                'conv_activation':self._conv_activation_vec[i], 
                'conv_activation_params':self._conv_activation_params_vec[i], 
                'norm_layer_position':self._conv_norm_layer_position_vec[i], 
                'norm_layer_type':self._conv_norm_layer_type_vec[i], 
                'norm_layer_params':self._conv_norm_layer_params_vec[i], 
                'pool_type':self._pool_type_vec[i], 
                'pool_kernel_size':self._pool_kernel_size_vec[i], 
                'pool_padding':self._pool_padding_vec[i], 
                'pool_stride':self._pool_stride_vec[i], 
                'pool_params':self._pool_params_vec[i], 
                'dropout':self._conv_dropout_vec[i], 
                'min_image_dim':self._min_image_size,
                'kernel_regularizer':(tf.keras.regularizers.L2(self._l2_reg) if self._l2_reg>0 else None)
            }
            if i==0:
                _kwargs.update({'input_shape':self._input_shape})
            d = add_conv_block(**_kwargs)
            self.net = d['model']
            output_image = d['output_image']
            self.size_list.append(output_image+[out_channels])
            in_channels = out_channels
            input_image = output_image
            
        # Flattening (Image embedding)
        self.net.add(tf.keras.layers.Flatten())
        self._dense_input_size = np.prod(output_image) * out_channels
        self.size_list.append([self._dense_input_size])
        
        # Dense layers hyperparameters
        self._dense_width = hparams["dense_width"]
        self._dense_depth = hparams["dense_depth"]
        self._dense_activation = hparams["dense_activation"] if hparams.get("dense_activation") else "relu"
        self._dense_activation_params = hparams.get("dense_activation_params")
        self._output_activation = hparams.get("output_activation") if hparams.get("output_activation") else None
        self._output_activation_params = hparams.get("output_activation_params")
        self._dense_norm_layer_type = hparams.get("dense_norm_layer_type")
        self._dense_norm_layer_params = hparams.get("dense_norm_layer_params")
        self._dense_norm_layer_position = hparams.get("dense_norm_layer_position")
        self._dense_dropout = hparams.get("dense_dropout")
        self._out_activation_module = actdict_keras[self._output_activation] if self._output_activation else None
        
        # Generate lists of hyperparameters for the dense layers
        self._dense_width_vec = self._gen_hparam_vec_for_dense(self._dense_width, 'dense_width',
            check_auto=True, init_for_auto=self._dense_input_size, powers_of_two_if_auto=True, direction_if_auto="down")
        self._dense_activation_vec = self._gen_hparam_vec_for_dense(self._dense_activation, 'dense_activation')
        self._dense_activation_params_vec = self._gen_hparam_vec_for_dense(self._dense_activation_params, 'dense_activation_params')
        self._dense_norm_layer_type_vec = self._gen_hparam_vec_for_dense(self._dense_norm_layer_type, 'dense_norm_layer_type')
        self._dense_norm_layer_params_vec = self._gen_hparam_vec_for_dense(self._dense_norm_layer_params, 'dense_norm_layer_params')
        self._dense_norm_layer_position_vec = self._gen_hparam_vec_for_dense(self._dense_norm_layer_position, 'dense_norm_layer_position')
        self._dense_dropout_vec = self._gen_hparam_vec_for_dense(self._dense_dropout, 'dense_dropout')
        
        # Construct the dense layers
        in_size = self._dense_input_size
        for i in range(self._dense_depth):
            out_size = self._dense_width_vec[i]
            temp_dropout_rate = self._dense_dropout_vec[i] if (i != self._dense_depth-1) else None # The hidden layer just before the output layer rarely has Dropout.
            _kwargs = {
                'model':self.net,
                'output_size':self._dense_width_vec[i],
                'activation':self._dense_activation_vec[i],
                'activation_params':self._dense_activation_params_vec[i],
                'norm_layer_type':self._dense_norm_layer_type_vec[i],
                'norm_layer_position':self._dense_norm_layer_position_vec[i],
                'norm_layer_params':self._dense_norm_layer_params_vec[i],
                'dropout':temp_dropout_rate,
                'kernel_regularizer':(tf.keras.regularizers.L2(self._l2_reg) if self._l2_reg>0 else None)
            }
            add_dense_block(**_kwargs)
            in_size = out_size
            self.size_list.append([out_size])
        
        # Output layer
        self.net.add(tf.keras.layers.Dense(self._output_shape[-1]))
        if self._output_activation:
            if self._output_activation_params: self.net.add(tf.keras.layers.Activation(self._out_activation_module(**self._output_activation_params)))
            else: self.net.add(tf.keras.layers.Activation(self._out_activation_module))

    
    def _gen_hparam_vec_for_conv(self, hparam, hparam_name, **kwargs):
        return generate_array_for_hparam(hparam, self._num_conv_blocks, hparam_name=hparam_name, count_if_not_list_name='num_conv_blocks', **kwargs)
    
    def _gen_hparam_vec_for_dense(self, hparam, hparam_name, **kwargs):
        return generate_array_for_hparam(hparam, self._dense_depth, hparam_name=hparam_name, count_if_not_list_name='dense_depth', **kwargs)
    
    def call(self, inputs, **kwargs):
        return self.net(inputs, **kwargs)
